/*
  GymBuddy - a Smart Fitness Device That Counts Reps and Tracks Your
  Progress in the Gym Using IMU sensor, we we're able to measure how many reps
  one does by counting the number of times a direction change is being made.
  We've integrated Blynk to be able to watch the number of reps in real time,
  and also control other features such as turning on LED fun light mode -
  for better mood during your workout, and fan control, directly from
  Blynk dashboard or by adding a siri shortcut.
  You can watch your results in a google sheet document, by sending an HTTP
  request to a designated Make.com URL.

  The circuit:
        LED strip       - IO15
        IR sender       - IO26
        Blue LED        - IO17
        Accelerometer:
          SCL           - IO22
          SDA           - IO21

  Video link: https://youtu.be/rm8J747FscE
  Created By:
        Michal Shapira
        Omri Nir
*/

// Blynk configurations
#define BLYNK_TEMPLATE_ID ""
#define BLYNK_TEMPLATE_NAME "GymBuddy"
#define BLYNK_AUTH_TOKEN ""

/* Comment this out to disable prints and save space */
#define BLYNK_PRINT Serial

#include <Arduino.h>
#include <BlynkSimpleEsp32.h>
#include <HTTPClient.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <WiFiClientSecure.h>
#include "I2Cdev.h" 
#include "MPU6050.h"
#include <Adafruit_NeoPixel.h>
#include <Adafruit_Sensor.h>

// Configuration Variables
// Your WiFi credentials.
// Set password to "" for open networks.
char ssid[] = "";
char pass[] = "";
// your Make.com URL
const char *makeUrl =
    "https://hook.eu2.make.com/MAKE.COM_URL?reps=";
// Variables for IR sender
uint8_t sCommand = 0x47;       // turn on fan
uint8_t sRepeats = 0;

// PINs
#define INDICATION_LED_PIN 17 // LED indication
#define IR_SEND_PIN 26        // IR SEND PIN
#define LED_PIN 15            // LED strip
#define LED_COUNT 12

// Sensor configurations
// IMU
MPU6050 accelgyro; // I2C address 0x68 by default

int16_t ax, ay, az;
int16_t gx, gy, gz;

#define OUTPUT_READABLE_ACCELGYRO
// IR Sender
#define DISABLE_CODE_FOR_RECEIVER // Disables restarting receiver after each
                                  // send. Saves 450 bytes program memory and
                                  // 269 bytes RAM if receiving functions are
                                  // not used.
// LED strip
Adafruit_NeoPixel strip(LED_COUNT, LED_PIN, NEO_GRB + NEO_KHZ800);

#include <IRremote.hpp>

// Variables for rep counting
float previousAccelY = 0;
int directionChanges = 0;
int reps = 0;
int lastSentReps = 0;
bool isInExercise = false;

// Variables for interval management
unsigned long lastSendTime = 0;
unsigned long sendInterval = 500; // Send data no more than every 1 second

// Variables for party mode
int funMode = 0;
unsigned long previousFunMillis = 0;
unsigned long funWait = 40;
long funFirstPixelHue = 0;

// Blynk virtual pins
// This function will be called every time Switch Widget - Start exercise
// in Blynk app writes values to the Virtual Pin V0
BLYNK_WRITE(V0) {
  int pinValue =
      param.asInt(); // assigning incoming value from pin V0 to a variable
  if (pinValue == 0) {
    // Send the number of reps to Make.com
    isInExercise = false;
    digitalWrite(INDICATION_LED_PIN,
                 LOW);         // turn the LED off by making the voltage LOW
    Blynk.virtualWrite(V2, 0); // Reset rep count in Virtual Pin V2
    report(reps);
  } else {
    isInExercise = true;
    digitalWrite(INDICATION_LED_PIN,
                 HIGH); // turn the LED on (HIGH is the voltage level)
  }
  reps = 0;
}

// This function will be called every time Switch Widget - Party mode
// in Blynk app writes values to the Virtual Pin V1
BLYNK_WRITE(V1) {
  funMode = param.asInt();
  if (funMode == 0) {
    clearLEDs();
  }
}

// This function will be called every time Switch Widget - Turn fan on
// in Blynk app writes values to the Virtual Pin V3
BLYNK_WRITE(V3) {
  param.asInt();
  Serial.println();
  Serial.print(F("Send now: address=0x00, command=0x"));
  Serial.print(sCommand, HEX);
  Serial.print(F(", repeats="));
  Serial.print(sRepeats);
  Serial.println();
  Serial.println(F("Send standard NEC with 8 bit address"));
  Serial.flush();
  Serial.println("Received IR request");
  IrSender.sendNEC(0x00, sCommand, sRepeats);
}

void setup() {
  // IMU
  Wire.begin();
  Serial.begin(115200);
  Serial.println("Initializing I2C devices...");
  accelgyro.initialize();
  Serial.println("Testing device connections...");
  Serial.println(accelgyro.testConnection() ? "MPU6050 connection successful"
                                            : "MPU6050 connection failed");

  // LED INDICATION
  pinMode(INDICATION_LED_PIN,
          OUTPUT); // initialize digital pin LED_BUILTIN as an output.

  // WiFi
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, pass);
  Serial.print("Waiting for WiFi to connect...");
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print(".");
    delay(1000);
  }
  Serial.println(" connected");

  // IR sender
  Serial.println(F("START " __FILE__ " from " __DATE__
                   "\r\nUsing library version " VERSION_IRREMOTE));
  Serial.print(F("Send IR signals at pin "));
  Serial.println(IR_SEND_PIN);
  IrSender.begin(); // Start with IR_SEND_PIN as send pin and if
                    // NO_LED_FEEDBACK_CODE is NOT defined, enable feedback LED
                    // at default feedback LED pin

  // Blynk
  Blynk.begin(BLYNK_AUTH_TOKEN, ssid, pass);

  Blynk.virtualWrite(V2, 0); // Reset rep count in Virtual Pin V2
}

void loop() {
  Blynk.run();

  run_fun_mode();

  // count reps if user has started exercise
  if (isInExercise && millis() - lastSendTime > sendInterval) {
    accelgyro.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);

    // Detect direction change
    if ((ay > 0 && previousAccelY < 0) || (ay < 0 && previousAccelY > 0)) {
      directionChanges++;
    }

    // Count reps
    if (directionChanges >= 2) {
      reps++;
      // Send data to Blynk at specified intervals
      directionChanges = 0;
    }

    // Send to Blynk with delay (make sure not to accidentaly send too much)
    if (reps != lastSentReps) {
      Serial.println("sending update to Blynk");
      Blynk.virtualWrite(V2, reps); // Send rep count to Virtual Pin V2
      lastSentReps = reps;
    }

    previousAccelY = ay;

    lastSendTime = millis();
  }
}

// HTTP functions
void report(int sendData) {
  Serial.println("starting report to make.com");
  WiFiClientSecure *client = new WiFiClientSecure;
  if (client) {
    client->setInsecure();
    {
      HTTPClient https;
      Serial.print("[HTTPS] begin...\n");
      Serial.print("Sending to ");
      Serial.println(String(makeUrl) + sendData);
      if (https.begin(*client, String(makeUrl) + sendData)) { // HTTPS
        Serial.print("[HTTPS] GET...\n");
        int httpCode = https.GET();
        if (httpCode > 0) {
          Serial.printf("[HTTPS] GET... code: %d\n", httpCode);
          if (httpCode == HTTP_CODE_OK ||
              httpCode == HTTP_CODE_MOVED_PERMANENTLY) {
            String payload = https.getString();
            Serial.println(payload);
          }
        } else {
          Serial.printf("[HTTPS] GET... failed, error: %s\n",
                        https.errorToString(httpCode).c_str());
        }
        https.end();
      } else {
        Serial.printf("[HTTPS] Unable to connect\n");
      }
    }
    delete client;
  } else {
    Serial.println("Unable to create client");
  }
  Serial.println();
}

// LED functions
void run_fun_mode() {
  unsigned long currentMillis = millis();
  if (funMode && currentMillis - previousFunMillis > funWait) {
    if (funFirstPixelHue >= 5 * 65536) {
      funFirstPixelHue = 0;
    }
    strip.rainbow(funFirstPixelHue);
    strip.show(); // Update strip with new contents
    funFirstPixelHue += 256;
  }
}

void clearLEDs() {
  for (int i = 0; i < strip.numPixels(); i++) {
    strip.setPixelColor(i, 0); // Set color to 'black' which is off
  }
  strip.show(); // Update the strip to turn off all LEDs
}
